#!/bin/bash
##
# Given a directory of snippets and a project's source code directory,
# find all occurrences of snippets within that project.
#
# Found occurrences are reported in csv-format. Each line reports
# (part of a) snippet and the parts of project files it was found in.
# Cells are separated by commas. The first line contains column headings.
#
# Approach:
#  * find clones within the snippets
#  * find clones within the project
#  * find clones within the union of snippets and project
#  * subtract the snippets-clones and the project-clones
#    from the clones of the union.
#  * what remains are occurrences of snippets within
#    the projects, or vice versa.



function main {
  # @todo cli usage description.
  #       check if necessary arguments are present.
  snippetsLocation="$1"
  projectLocation="$2"
  snippetsCache="$3"
  projectCache="$4"
  arguments="$5"
  # name of the function to convert one line of PMD CPD's output
  outputFormat="$6"
  # set detection arguments, given by user

  DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
  # set large heap size to avoid java.lang.OutOfMemoryError exception
  export HEAPSIZE=4096m
  # @todo test if binary exists
  binCpd="$DIR/ccd/pmd-bin-5.4.1/bin/run.sh cpd --language java --encoding utf-8 --skip-lexical-errors --format csv_with_linecount_per_file"


  # find snippets baseline clones
  # @todo take data from the cache, if already exists
#  snippetsBaseline="${snippetsCache}/snippets-clones.txt"
#  mkdir -p `dirname "$snippetsBaseline"`
#  echo $binCpd $arguments --files "$snippetsLocation" | deflate > "$snippetsBaseline"


  # find project baseline clones
#  projectBaseline="${projectCache}/project-clones.txt"
#  mkdir -p `dirname "$projectBaseline"`
#  $binCpd $arguments --files "$projectLocation" | deflate > "$projectBaseline"


  # find clones in the project-snippets union
  projectSnippetsClones="${projectCache}/union-clones.txt"
  mkdir -p `dirname "$projectSnippetsClones"`
  $binCpd $arguments --files "$snippetsLocation" --files "$projectLocation" | deflate > "$projectSnippetsClones"

  projectSnippetsIntersection="${projectCache}/intersection-clones.txt"
  
  # remove clone-pairs within snippets and those within project-files.
  # For escaping cf. http://stackoverflow.com/a/2705678
  escapedSnippetsLocation=$(sed -e 's/[]\/$*.^|[]/\\&/g' <<< "$snippetsLocation")
  escapedProjectLocation=$(sed -e 's/[]\/$*.^|[]/\\&/g' <<< "$projectLocation")

  snippetsClonePattern="/^[0-9]\+,[0-9]\+,$escapedSnippetsLocation[^,]\+,[0-9]\+,[0-9]\+,$escapedSnippetsLocation[^,]\+$/d"
  projectClonePattern="/^[0-9]\+,[0-9]\+,$escapedProjectLocation[^,]\+,[0-9]\+,[0-9]\+,$escapedProjectLocation[^,]\+$/d"

  sed -e "$snippetsClonePattern" "$projectSnippetsClones" | sed -e "$projectClonePattern" > "$projectSnippetsIntersection"

  # subtract snippets-clones and project-clones from project-snippets clones
#  unionMinusProjectClones="${projectCache}/union-minus-project-clones.txt"
#  projectSnippetsIntersection="${projectCache}/intersection-clones.txt"
#  # cf. http://unix.stackexchange.com/a/158235 for the following line
#  # @todo if speed up is needed, try comm with sorted files, see http://unix.stackexchange.com/a/158239
#  # @fixme somehow it's wrong!!
#  grep -vxFf "$projectBaseline" "$projectSnippetsClones" > "$unionMinusProjectClones"
#
#  grep -vxFf "$snippetsBaseline" "$unionMinusProjectClones" > "$projectSnippetsIntersection"

  
  # report remaining clones as occurrences of snippets within project
  ${outputFormat} "$projectSnippetsIntersection"
}


function deflate {

  while read line; do
    if [[ -z "${line// }" ]] # line is empty, cf. http://stackoverflow.com/a/918931
  	then
      continue
    fi

  	IFS=',' read -ra fields <<< $line # parse comma separated line,
                                      # cf. http://stackoverflow.com/a/918931

    if [ ${#fields[@]} -lt 8 ] # skips header, for example
    then
      continue
    fi

    # parsing PMD CPD's csv_with_linecount_per_file format.
    # Best to see their implementation for details at
    # https://github.com/pmd/pmd/blob/master/pmd-core/src/main/java/net/sourceforge/pmd/cpd/CSVRenderer.java
    numberOfClones=${fields[1]}

    # @todo make sure that the snippet is named first.
    #       Might already be ensured due to the order of --files arguments
    #       of the CPD binary. Did not see anything to the contrary, up to now.
    for i in `seq 1 $((numberOfClones - 1))`;
    do
      aStart=${fields[$i*3-1]}
      aSize=${fields[$i*3-1 +1]}
      aFile=${fields[$i*3-1 +2]}
      
      for j in `seq $((i + 1)) $numberOfClones`;
      do
        bStart=${fields[$j*3-1]}
        bSize=${fields[$j*3-1 +1]}
        bFile=${fields[$j*3-1 +2]}

        echo "$aStart,$aSize,$aFile,$bStart,$bSize,$bFile"
      done
    done

  done

}


# Output format which only reports the relation that a clone
# was found between what project file and what snippet file.
# Format is:
# snippet-file-path,project-file-path
function outputFormatRelation {
  while read line; do
    _convert_line_relation "$line"
  done < "$1"
}

function _convert_line_relation {
    line="$1"
    if [[ -z "${line// }" ]] # line is empty, cf. http://stackoverflow.com/a/918931
  	then
      return
    fi

  	IFS=',' read -ra fields <<< $line # parse comma separated line,
                                      # cf. http://stackoverflow.com/a/918931

    echo "${fields[2]},${fields[5]}"
}


# Output format optimized for visual inspection, i.e. it contains
# sourcecode excerpts from the snippet or project files
function outputFormatVisual {
  echo -n "snippet_start,snippet_end,snippet_file,snippet_code,"
  echo "occurrence_start,occurrence_end,occurrence_file,occurrence_code"
  while read line; do
    _convert_line "$line"
  done < "$1"
}

##
# Add the source code of detected clones.
function _convert_line {
    line="$1"
    if [[ -z "${line// }" ]] # line is empty, cf. http://stackoverflow.com/a/918931
  	then
      return
    fi

  	IFS=',' read -ra fields <<< $line # parse comma separated line,
                                      # cf. http://stackoverflow.com/a/918931
    # parsing PMD CPD's csv_with_linecount_per_file format.
    # Best to see their implementation for details at
    # https://github.com/pmd/pmd/blob/master/pmd-core/src/main/java/net/sourceforge/pmd/cpd/CSVRenderer.java
    # @todo make sure that the snippet is named first.
    #       Might already be ensured due to the order of --files arguments
    #       of the CPD binary. Did not see anything to the contrary, up to now.
    for i in 0 1;
    do
      cloneStart=${fields[$i*3]}
      cloneSize=${fields[$i*3 +1]}
      cloneFile=${fields[$i*3 +2]}
      cloneStart=$((cloneStart + 1)) # PMD starts counting at 0
      cloneEnd=$((cloneStart + cloneSize -1))
      echo -n "$cloneStart,$cloneEnd,$cloneFile,"
      echo -n '"'
      sed -n "$cloneStart,${cloneEnd}p" $cloneFile | sed -e 's/\r//g' | sed -e 's/"/""/g'
      echo -n '"'
      if [ "$i" -eq "0" ]
      then
        echo -n ','
      fi
    done
    echo
}

main "$@"
